@@ -14,92 +14,112 @@
 
 #include "test-utils.h"
 
+char *server2_uri;
+
 typedef struct {
 	const char *method;
 	const char *path;
 	guint status_code;
+	gboolean repeat;
 } TestRequest;
 
 static struct {
 	TestRequest requests[3];
+	guint final_status;
 } tests[] = {
 	/* A redirecty response to a GET or HEAD should cause a redirect */
 
 	{ { { "GET", "/301", 301 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "GET", "/302", 302 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "GET", "/303", 303 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "GET", "/307", 307 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "HEAD", "/301", 301 },
 	    { "HEAD", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "HEAD", "/302", 302 },
 	    { "HEAD", "/", 200 },
-	    { NULL } } },
-	/* 303 is a nonsensical response to HEAD, so we don't care
-	 * what happens there.
+	    { NULL } }, 200 },
+	/* 303 is a nonsensical response to HEAD, but some sites do
+	 * it anyway. :-/
 	 */
+	{ { { "HEAD", "/303", 303 },
+	    { "HEAD", "/", 200 },
+	    { NULL } }, 200 },
 	{ { { "HEAD", "/307", 307 },
 	    { "HEAD", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 
 	/* A non-redirecty response to a GET or HEAD should not */
 
 	{ { { "GET", "/300", 300 },
-	    { NULL } } },
+	    { NULL } }, 300 },
 	{ { { "GET", "/304", 304 },
-	    { NULL } } },
+	    { NULL } }, 304 },
 	{ { { "GET", "/305", 305 },
-	    { NULL } } },
+	    { NULL } }, 305 },
 	{ { { "GET", "/306", 306 },
-	    { NULL } } },
+	    { NULL } }, 306 },
 	{ { { "GET", "/308", 308 },
-	    { NULL } } },
+	    { NULL } }, 308 },
 	{ { { "HEAD", "/300", 300 },
-	    { NULL } } },
+	    { NULL } }, 300 },
 	{ { { "HEAD", "/304", 304 },
-	    { NULL } } },
+	    { NULL } }, 304 },
 	{ { { "HEAD", "/305", 305 },
-	    { NULL } } },
+	    { NULL } }, 305 },
 	{ { { "HEAD", "/306", 306 },
-	    { NULL } } },
+	    { NULL } }, 306 },
 	{ { { "HEAD", "/308", 308 },
-	    { NULL } } },
+	    { NULL } }, 308 },
 	
 	/* Test double-redirect */
 
 	{ { { "GET", "/301/302", 301 },
 	    { "GET", "/302", 302 },
-	    { "GET", "/", 200 } } },
+	    { "GET", "/", 200 } }, 200 },
 	{ { { "HEAD", "/301/302", 301 },
 	    { "HEAD", "/302", 302 },
-	    { "HEAD", "/", 200 } } },
+	    { "HEAD", "/", 200 } }, 200 },
 
-	/* POST should only automatically redirect on 302 and 303 */
+	/* POST should only automatically redirect on 301, 302 and 303 */
 
 	{ { { "POST", "/301", 301 },
-	    { NULL } } },
+	    { "GET", "/", 200 },
+	    { NULL } }, 200 },
 	{ { { "POST", "/302", 302 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "POST", "/303", 303 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "POST", "/307", 307 },
-	    { NULL } } },
+	    { NULL } }, 307 },
 
-	/* Test behavior with recoverably-bad Location header
-	 */
+	/* Test behavior with recoverably-bad Location header */
 	{ { { "GET", "/bad", 302 },
 	    { "GET", "/bad%20with%20spaces", 200 },
-	    { NULL } } }
+	    { NULL } }, 200 },
+
+	/* Test behavior with irrecoverably-bad Location header */
+	{ { { "GET", "/bad-no-host", 302 },
+	    { NULL } }, SOUP_STATUS_MALFORMED },
+
+	/* Test infinite redirection */
+	{ { { "GET", "/bad-recursive", 302, TRUE },
+	    { NULL } }, SOUP_STATUS_TOO_MANY_REDIRECTS },
+
+	/* Test redirection to a different server */
+	{ { { "GET", "/server2", 302 },
+	    { "GET", "/on-server2", 200 },
+	    { NULL } }, 200 },
 };
 static const int n_tests = G_N_ELEMENTS (tests);
 
@@ -134,7 +154,7 @@ restarted (SoupMessage *msg, gpointer user_data)
 
 	debug_printf (2, "    %s %s\n", msg->method, uri->path);
 
-	if ((*req)->method)
+	if ((*req)->method && !(*req)->repeat)
 		(*req)++;
 
 	if (!(*req)->method) {
@@ -182,6 +202,13 @@ do_test (SoupSession *session, SoupURI *base_uri, int n)
 			  G_CALLBACK (restarted), &req);
 
 	soup_session_send_message (session, msg);
+
+	if (msg->status_code != tests[n].final_status) {
+		debug_printf (1, "    - Expected final status of %d, got %d !\n",
+			      tests[n].final_status, msg->status_code);
+		errors++;
+	}
+
 	g_object_unref (msg);
 	debug_printf (2, "\n");
 }
@@ -219,14 +246,28 @@ server_callback (SoupServer *server, SoupMessage *msg,
 			soup_message_headers_replace (msg->response_headers,
 						      "Location",
 						      "/bad with spaces");
+		} else if (!strcmp (path, "/bad-recursive")) {
+			soup_message_set_status (msg, SOUP_STATUS_FOUND);
+			soup_message_headers_replace (msg->response_headers,
+						      "Location",
+						      "/bad-recursive");
+		} else if (!strcmp (path, "/bad-no-host")) {
+			soup_message_set_status (msg, SOUP_STATUS_FOUND);
+			soup_message_headers_replace (msg->response_headers,
+						      "Location",
+						      "about:blank");
 		} else if (!strcmp (path, "/bad with spaces"))
 			soup_message_set_status (msg, SOUP_STATUS_OK);
 		else
 			soup_message_set_status (msg, SOUP_STATUS_NOT_FOUND);
 		return;
-	}
-
-	if (!strcmp (path, "/")) {
+	} else if (!strcmp (path, "/server2")) {
+		soup_message_set_status (msg, SOUP_STATUS_FOUND);
+		soup_message_headers_replace (msg->response_headers,
+					      "Location",
+					      server2_uri);
+		return;
+	} else if (!strcmp (path, "/")) {
 		if (msg->method != SOUP_METHOD_GET &&
 		    msg->method != SOUP_METHOD_HEAD) {
 			soup_message_set_status (msg, SOUP_STATUS_METHOD_NOT_ALLOWED);
@@ -282,6 +323,14 @@ server_callback (SoupServer *server, SoupMessage *msg,
 	}
 }
 
+static void
+server2_callback (SoupServer *server, SoupMessage *msg,
+		  const char *path, GHashTable *query,
+		  SoupClientContext *context, gpointer data)
+{
+	soup_message_set_status (msg, SOUP_STATUS_OK);
+}
+
 static gboolean run_tests = TRUE;
 
 static GOptionEntry no_test_entry[] = {
@@ -295,7 +344,7 @@ int
 main (int argc, char **argv)
 {
 	GMainLoop *loop;
-	SoupServer *server;
+	SoupServer *server, *server2;
 	guint port;
 	SoupURI *base_uri;
 
@@ -304,7 +353,13 @@ main (int argc, char **argv)
 	server = soup_test_server_new (TRUE);
 	soup_server_add_handler (server, NULL,
 				 server_callback, NULL, NULL);
-	port = 	soup_server_get_port (server);
+	port = soup_server_get_port (server);
+
+	server2 = soup_test_server_new (TRUE);
+	soup_server_add_handler (server2, NULL,
+				 server2_callback, NULL, NULL);
+	server2_uri = g_strdup_printf ("http://127.0.0.1:%d/on-server2",
+				       soup_server_get_port (server2));
 
 	loop = g_main_loop_new (NULL, TRUE);
 
@@ -319,6 +374,9 @@ main (int argc, char **argv)
 	}
 
 	g_main_loop_unref (loop);
+	g_free (server2_uri);
+	soup_test_server_quit_unref (server);
+	soup_test_server_quit_unref (server2);
 
 	if (run_tests)
 		test_cleanup ();
